<!DOCTYPE html>
<html lang="en">

<head>
    <meta charset="UTF-8">
    <title>{{ title|default:"DBMI App" }} Login</title>

    {% load static %}
    {% load bootstrap3 %}

    <link rel="shortcut icon" sizes="16x16 24x24 32x32 48x48" href="{% static 'dbmi_client/login/favicon/favicon.ico' %}">
    <link rel="apple-touch-icon" sizes="180x180" href="{% static 'dbmi_client/login/favicon/apple-touch-icon.png' %}">
    <link rel="icon" type="image/png" sizes="32x32" href="{% static 'dbmi_client/login/favicon/favicon-32x32.png' %}">
    <link rel="icon" type="image/png" sizes="16x16" href="{% static 'dbmi_client/login/favicon/favicon-16x16.png' %}">
    <link rel="manifest" href="{% static 'dbmi_client/login/favicon/site.webmanifest' %}">
    <link rel="mask-icon" href="{% static 'dbmi_client/login/favicon/safari-pinned-tab.svg' %}" color="#5bbad5">
    <meta name="theme-color" content="#ffffff">
    <meta name="msapplication-config" content="{% static 'dbmi_client/login/favicon/browserconfig.xml' %}">
    <meta name="msapplication-TileImage" content="{% static 'dbmi_client/login/favicon/mstile-large.png' %}">
    <meta name="msapplication-TileColor" content="#da532c">

    {% bootstrap_css %}

    <link rel="stylesheet" type="text/css" href="{% static 'css/main.css' %}" />
    {% block stylesheets %}{% endblock %}

</head>
    <body>

        <nav id="nav-bar" class="navbar navbar-default">
            <div class="container-fluid">
                <div class="row">
                    <div class="center-block">
                        <div class="navbar-header">
                            <a class="navbar-brand" href="https://dbmi.hms.harvard.edu"><img height="70" src="{% static 'dbmi_client/login/img/hms_dbmi_logo.png' %}" /></a>
                        </div>
                    </div>
                </div>
            </div>
        </nav>

        <div id="container" class="container">
            {% block content %}{% endblock %}
        </div>

        <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.12.4/jquery.min.js"></script>
        {% bootstrap_javascript %}
        {% block javascript %}{% endblock %}
    </body>
</html>
